#pragma once

#include <iostream>
#include <string_view>
#include <source_location>
#include <cxxabi.h>
#include <libunwind.h>
#include <dlfcn.h>
#include <cstring>
#include <memory>
#include <sstream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>

// Simple RAII wrapper for file descriptor
class FileDescriptor {
    int fd_ = -1;
public:
    explicit FileDescriptor(int fd) : fd_(fd) {}
    ~FileDescriptor() { if (fd_ != -1) close(fd_); }
    operator int() const { return fd_; }
    FileDescriptor(const FileDescriptor&) = delete;
    FileDescriptor& operator=(const FileDescriptor&) = delete;
};

std::string get_addr2line_info(const char* executable, void* addr) {
    char cmd[512];
    snprintf(cmd, sizeof(cmd), "addr2line -f -p -e %s %p 2>/dev/null", executable, addr);
    
    FILE* pipe = popen(cmd, "r");
    if (!pipe) return "";
    
    char buffer[1024];
    std::string result;
    while (fgets(buffer, sizeof(buffer), pipe) != nullptr) {
        result += buffer;
    }
    pclose(pipe);
    
    // Remove trailing newline if present
    if (!result.empty() && result.back() == '\n') {
        result.pop_back();
    }
    return result;
}

void print_stacktrace() {
    unw_cursor_t cursor;
    unw_context_t context;
    unw_word_t offset, pc;
    char sym[256];
    char *name = sym;
    int status;
    int frame_num = 0;
    char exe_path[1024] = {0};

    // Get the path to the current executable
    ssize_t count = readlink("/proc/self/exe", exe_path, sizeof(exe_path) - 1);
    if (count == -1) {
        strcpy(exe_path, "[unknown]");
    } else {
        exe_path[count] = '\0';
    }

    // Initialize cursor to current frame
    unw_getcontext(&context);
    unw_init_local(&cursor, &context);

    std::cerr << "Stack trace (most recent call first):\n";

    // Walk up the stack
    while (unw_step(&cursor) > 0) {
        unw_get_reg(&cursor, UNW_REG_IP, &pc);
        if (pc == 0) {
            break;
        }

        std::cerr << "#" << frame_num++ << " ";
        
        // Get symbol name
        if (unw_get_proc_name(&cursor, sym, sizeof(sym), &offset) == 0) {
            // Demangle the C++ name
            int status = 0;
            size_t length = sizeof(sym);
            name = abi::__cxa_demangle(sym, nullptr, &length, &status);
            
            // Get line info using addr2line
            std::string line_info = get_addr2line_info(exe_path, (void*)pc);
            
            Dl_info info;
            if (dladdr(reinterpret_cast<void*>(pc), &info) && info.dli_fname) {
                const char* base = strrchr(info.dli_fname, '/');
                std::cerr << (base ? base + 1 : info.dli_fname) << ": ";
            }
            
            if (name) {
                std::cerr << name;
                free(name);
            } else {
                std::cerr << sym;
            }
            
            std::cerr << " + 0x" << std::hex << offset;
            
            if (!line_info.empty()) {
                std::cerr << " at " << line_info;
            }
            
            std::cerr << " [0x" << std::hex << pc << "]\n";
        } else {
            std::cerr << "[unknown] [0x" << std::hex << pc << "]";
            // Try to get at least some info with addr2line even for unknown symbols
            std::string line_info = get_addr2line_info(exe_path, (void*)pc);
            if (!line_info.empty()) {
                std::cerr << " at " << line_info;
            }
            std::cerr << "\n";
        }
    }
}

void assert_failed(
    bool condition,
    std::string_view message,
    std::source_location location = std::source_location::current()
) {
    if (!condition) {
        std::cerr << "Assertion failed at " << location.file_name() << ":" << location.line() << ": " 
                  << location.function_name() << "(): " << message << "\n";
        print_stacktrace();
        std::abort();
    }
}

#define ASSERT(condition, message) assert_failed(condition, message, std::source_location::current())
